# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
SONATA network config parsing.
"""

# TODO: move to `libsonata` library

import collections
import os.path

import six

from bluepysnap import utils


class Config(object):
    """
    SONATA network config parser.

    See also:
        https://github.com/AllenInstitute/sonata/blob/master/docs/SONATA_DEVELOPER_GUIDE.md#network_config
    """
    def __init__(self, filepath):
        configdir = os.path.abspath(os.path.dirname(filepath))
        content = utils.load_json(filepath)
        self.manifest = Config._resolve_manifest(content.pop('manifest'), configdir)
        self.content = content

    @staticmethod
    def _resolve_manifest(manifest, configdir):
        result = manifest.copy()

        assert '${configdir}' not in result
        result['${configdir}'] = configdir

        for k, v in six.iteritems(result):
            if v == '.':
                result[k] = configdir
            elif v.startswith('./'):
                result[k] = os.path.join(configdir, v[2:])

        while True:
            update = False
            for k, v in six.iteritems(result):
                if v.startswith('$'):
                    tokens = v.split('/', 1)
                    resolved = result[tokens[0]]
                    if '$' not in resolved:
                        result[k] = os.path.join(resolved, *tokens[1:])
                        update = True
            if not update:
                break

        return result

    def _resolve_path(self, value):
        if '$' in value:
            vs = [
                self.manifest[v] if v.startswith('$') else v
                for v in value.split('/')
            ]
            return os.path.join(*vs)
        else:
            return value

    def _resolve(self, value):
        if isinstance(value, collections.Mapping):
            return {
                k: self._resolve(v) for k, v in six.iteritems(value)
            }
        elif isinstance(value, six.string_types):
            return self._resolve_path(value)
        elif isinstance(value, collections.Iterable):
            return [self._resolve(v) for v in value]
        else:
            return value

    def resolve(self):
        """ Resolve variables in config file paths. """
        return self._resolve(self.content)

    @staticmethod
    def parse(filepath):
        """ Parse SONATA network config. """
        return Config(filepath).resolve()



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Access to circuit data.
"""

from cached_property import cached_property

from bluepysnap.config import Config
from bluepysnap.nodes import NodePopulation
from bluepysnap.edges import EdgePopulation
from bluepysnap.exceptions import BlueSnapError


def _collect_populations(configs, cls, select=None):
    result = {}
    for cfg in configs:
        population = cls(cfg)
        if population.name in result:
            raise BlueSnapError("Duplicate population: '%s'" % population.name)
        result[population.name] = population
    if select is None:
        return result
    elif select in result:
        return result[select]
    else:
        raise BlueSnapError("No such population: '%s'" % select)


class Circuit(object):
    """ Access to circuit data. """
    def __init__(self, config, node_population=None, edge_population=None):
        self._config = Config(config).resolve()
        self._node_population = node_population
        self._edge_population = edge_population

    @property
    def config(self):
        """ Network config dictionary. """
        return self._config

    @cached_property
    def nodes(self):
        """ Access to node population(s). """
        return _collect_populations(
            self._config['networks']['nodes'],
            lambda cfg: NodePopulation(cfg, self),
            select=self._node_population
        )

    @cached_property
    def edges(self):
        """ Access to edge population(s). """
        return _collect_populations(
            self._config['networks']['edges'],
            lambda cfg: EdgePopulation(cfg, self),
            select=self._edge_population
        )



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Edge population access.
"""

from builtins import map

import collections

import libsonata
import numpy as np
import pandas as pd

from cached_property import cached_property

from bluepysnap.exceptions import BlueSnapError
from bluepysnap.utils import is_iterable

SOURCE_NODE_ID = "@source_node"
TARGET_NODE_ID = "@target_node"
DYNAMICS_PREFIX = "@dynamics:"


def _get_population_name(h5_filepath):
    populations = libsonata.EdgeStorage(h5_filepath).population_names
    if len(populations) != 1:
        raise BlueSnapError(
            "Only single-population node collections are supported (found: %d)" % len(populations)
        )
    return list(populations)[0]


def _resolve_node_ids(nodes, group):
    """ Node IDs corresponding to node group filter. """
    if group is None:
        return None
    return nodes.ids(group)


def _is_empty(xs):
    return (xs is not None) and (len(xs) == 0)


def _estimate_range_size(func, node_ids, n=3):
    """ Median size of index second level for some node IDs from the provided list. """
    assert len(node_ids) > 0
    if len(node_ids) > n:
        node_ids = np.random.choice(node_ids, size=n, replace=False)
    return np.median([
        len(func(node_id).ranges) for node_id in node_ids
    ])


class EdgePopulation(object):
    """ Edge population access. """

    def __init__(self, config, circuit):
        self._h5_filepath = config['edges_file']
        self._csv_filepath = config['edge_types_file']
        self._circuit = circuit

    @cached_property
    def _population(self):
        return libsonata.EdgePopulation(self._h5_filepath, self._csv_filepath or '', self.name)

    @cached_property
    def name(self):
        """ Population name. """
        return _get_population_name(self._h5_filepath)

    @property
    def size(self):
        """ Population size. """
        return self._population.size

    def _nodes(self, population):
        nodes = self._circuit.nodes

        result = None
        if isinstance(nodes, collections.Mapping):
            result = nodes.get(population)
        elif nodes.name == population:
            result = nodes

        if result is None:
            raise BlueSnapError("Undefined node population: '%s'" % population)

        return result

    @property
    def source(self):
        """ Source NodePopulation. """
        return self._nodes(self._population.source)

    @cached_property
    def target(self):
        """ Target NodePopulation. """
        return self._nodes(self._population.target)

    @cached_property
    def _property_names(self):
        return set(self._population.attribute_names)

    @cached_property
    def _dynamics_params_names(self):
        return {DYNAMICS_PREFIX + name for name in list(self._population.dynamics_attribute_names)}

    @property
    def property_names(self):
        """ Set of available edge properties. """
        return self._property_names | self._dynamics_params_names

    def _get_property(self, prop, selection):
        if prop == SOURCE_NODE_ID:
            result = self._population.source_nodes(selection)
        elif prop == TARGET_NODE_ID:
            result = self._population.target_nodes(selection)
        elif prop in self._property_names:
            result = self._population.get_attribute(prop, selection)
        elif prop in self._dynamics_params_names:
            result = self._population.get_dynamics_attribute(
                prop.split(DYNAMICS_PREFIX)[1], selection)
        else:
            raise BlueSnapError("No such property: %s" % prop)
        return result

    def _get(self, selection, properties=None):
        """ Get an array of edge IDs or DataFrame with edge properties. """
        edge_ids = selection.flatten()

        if properties is None:
            return edge_ids

        if is_iterable(properties):
            if len(edge_ids) == 0:
                result = pd.DataFrame(columns=properties)
            else:
                result = pd.DataFrame(index=edge_ids)
                for p in properties:
                    result[p] = self._get_property(p, selection)
        else:
            if len(edge_ids) == 0:
                result = pd.Series(name=properties)
            else:
                result = pd.Series(
                    self._get_property(properties, selection),
                    index=edge_ids,
                    name=properties
                )

        return result

    def properties(self, edge_ids, properties):
        """
        Edge properties as pandas DataFrame.

        Args:
            edge_ids: array-like of edge IDs
            properties: edge property name | list of edge property names

        Returns:
            Pandas Series indexed by edge IDs if `properties` is scalar;
            Pandas DataFrame indexed by edge IDs if `properties` is list.
        """
        selection = libsonata.Selection(edge_ids)
        return self._get(selection, properties)

    def positions(self, edge_ids, side, kind):
        """
        Edge positions as pandas DataFrame.

        Args:
            edge_ids: array-like of edge IDs
            side: 'afferent' | 'efferent'
            kind: 'center' | 'surface'

        Returns:
            Pandas Dataframe with ('x', 'y', 'z') columns indexed by edge IDs.
        """
        assert side in ('afferent', 'efferent')
        assert kind in ('center', 'surface')
        props = {
            '{side}_{kind}_{p}'.format(side=side, kind=kind, p=p): p
            for p in ['x', 'y', 'z']
        }
        result = self.properties(edge_ids, list(props))
        result.rename(columns=props, inplace=True)
        result.sort_index(axis=1, inplace=True)
        return result

    def afferent_nodes(self, node_id, unique=True):
        """
        Get afferent node IDs for given target `node_id`.

        Args:
            node_id: target node ID

        Returns:
            Array of source node IDs.
            By default, the array is uniq-ed and sorted.
        """
        selection = self._population.afferent_edges(
            _resolve_node_ids(self.target, node_id)
        )
        result = self._population.source_nodes(selection)
        if unique:
            result = np.unique(result)
        return result

    def efferent_nodes(self, node_id, unique=True):
        """
        Get efferent node IDs for given source `node_id`.

        Args:
            node_id: source node ID

        Returns:
            Array of target node IDs.
            By default, the array is uniq-ed and sorted.
        """
        selection = self._population.efferent_edges(
            _resolve_node_ids(self.source, node_id)
        )
        result = self._population.target_nodes(selection)
        if unique:
            result = np.unique(result)
        return result

    def afferent_edges(self, node_id, properties=None):
        """
        Get afferent edges for given `node_id`.

        Args:
            node_id: target node ID
            properties: None / edge property name / list of edge property names

        Returns:
            List of edge IDs, if `properties` is None;
            Pandas Series indexed by edge IDs if `properties` is string;
            Pandas DataFrame indexed by edge IDs if `properties` is list.
        """
        return self.pathway_edges(source=None, target=node_id, properties=properties)

    def efferent_edges(self, node_id, properties=None):
        """
        Get efferent edges for given `node_id`.

        Args:
            node_id: source node ID
            properties: None / edge property name / list of edge property names

        Returns:
            List of edge IDs, if `properties` is None;
            Pandas Series indexed by edge IDs if `properties` is string;
            Pandas DataFrame indexed by edge IDs if `properties` is list.
        """
        return self.pathway_edges(source=node_id, target=None, properties=properties)

    def pair_edges(self, source_node_id, target_node_id, properties=None):
        """
        Get edges corresponding to `source_node_id` -> `target_node_id` connection.

        Args:
            source_node_id: source node ID
            target_node_id: target node ID
            properties: None / edge property name / list of edge property names

        Returns:
            List of edge IDs, if `properties` is None;
            Pandas Series indexed by edge IDs if `properties` is string;
            Pandas DataFrame indexed by edge IDs if `properties` is list.
        """
        return self.pathway_edges(
            source=source_node_id, target=target_node_id, properties=properties
        )

    def pathway_edges(self, source=None, target=None, properties=None):
        """
        Get edges corresponding to `source` -> `target` connections.

        Args:
            source: source node group
            target: target node group
            properties: None / edge property name / list of edge property names

        Returns:
            List of edge IDs, if `properties` is None;
            Pandas Series indexed by edge IDs if `properties` is string;
            Pandas DataFrame indexed by edge IDs if `properties` is list.
        """
        if source is None and target is None:
            raise BlueSnapError("Either `source` or `target` should be specified")

        source_node_ids = _resolve_node_ids(self.source, source)
        target_edge_ids = _resolve_node_ids(self.target, target)

        if source_node_ids is None:
            selection = self._population.afferent_edges(target_edge_ids)
        elif target_edge_ids is None:
            selection = self._population.efferent_edges(source_node_ids)
        else:
            selection = self._population.connecting_edges(source_node_ids, target_edge_ids)

        return self._get(selection, properties)

    def _iter_connections(self, source_node_ids, target_node_ids, unique_node_ids, shuffle):
        """ Iterate through `source_node_ids` -> `target_node_ids` connections. """

        # pylint: disable=too-many-branches,too-many-locals
        def _optimal_direction():
            """ Choose between source and target node IDs for iterating. """
            if target_node_ids is None and source_node_ids is None:
                raise BlueSnapError("Either `source` or `target` should be specified")
            if source_node_ids is None:
                return 'target'
            if target_node_ids is None:
                return 'source'
            else:
                range_size_source = _estimate_range_size(
                    self._population.efferent_edges, source_node_ids
                )
                range_size_target = _estimate_range_size(
                    self._population.afferent_edges, target_node_ids
                )
                return 'source' if (range_size_source < range_size_target) else 'target'

        if _is_empty(source_node_ids) or _is_empty(target_node_ids):
            return

        direction = _optimal_direction()
        if direction == 'target':
            primary_node_ids, secondary_node_ids = target_node_ids, source_node_ids
            get_connected_node_ids = self.afferent_nodes
        else:
            primary_node_ids, secondary_node_ids = source_node_ids, target_node_ids
            get_connected_node_ids = self.efferent_nodes

        primary_node_ids = np.unique(primary_node_ids)
        if shuffle:
            np.random.shuffle(primary_node_ids)

        if secondary_node_ids is not None:
            secondary_node_ids = np.unique(secondary_node_ids)

        secondary_node_ids_used = set()

        for key_node_id in primary_node_ids:
            connected_node_ids = get_connected_node_ids(key_node_id, unique=False)
            connected_node_ids_with_count = np.stack(
                np.unique(connected_node_ids, return_counts=True)
            ).transpose()
            # np.stack(uint64, int64) -> float64
            connected_node_ids_with_count = connected_node_ids_with_count.astype(np.uint32)
            if secondary_node_ids is not None:
                mask = np.in1d(connected_node_ids_with_count[:, 0],
                               secondary_node_ids, assume_unique=True)
                connected_node_ids_with_count = connected_node_ids_with_count[mask]
            if shuffle:
                np.random.shuffle(connected_node_ids_with_count)
            for conn_node_id, edge_count in connected_node_ids_with_count:
                if direction == 'target':
                    yield conn_node_id, key_node_id, edge_count
                else:
                    yield key_node_id, conn_node_id, edge_count
                if unique_node_ids:
                    secondary_node_ids_used.add(conn_node_id)
                    break

    def iter_connections(
            self, source=None, target=None, unique_node_ids=False, shuffle=False,
            return_edge_ids=False, return_edge_count=False
    ):
        """
        Iterate through `source` -> `target` connections.

        Args:
            source: source node group
            target: target node group
            unique_node_ids: if True, no node ID would be used more than once
            shuffle: if True, result order would be (somewhat) randomised
            return_edge_count: if True, edge count is added to yield result
            return_edge_ids: if True, edge ID list is added to yield result

        `return_edge_count` and `return_edge_ids` are mutually exclusive.

        Yields:
            (source_node_id, target_node_id, edge_ids) if return_edge_ids == True;
            (source_node_id, target_node_id, edge_count) if return_edge_count == True;
            (source_node_id, target_node_id) otherwise.
        """
        if return_edge_ids and return_edge_count:
            raise BlueSnapError(
                "`return_edge_count` and `return_edge_ids` are mutually exclusive"
            )

        source_node_ids = _resolve_node_ids(self.source, source)
        target_node_ids = _resolve_node_ids(self.target, target)

        it = self._iter_connections(source_node_ids, target_node_ids, unique_node_ids, shuffle)

        if return_edge_count:
            return it
        elif return_edge_ids:
            add_edge_ids = lambda x: (x[0], x[1], self.pair_edges(x[0], x[1]))
            return map(add_edge_ids, it)
        else:
            omit_edge_count = lambda x: x[:2]
            return map(omit_edge_count, it)



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Morphology access.
"""

import os

import numpy as np
import neurom as nm

from bluepysnap.settings import MORPH_CACHE_SIZE


class MorphHelper(object):
    """ Collection of morphology-related methods. """

    def __init__(self, morph_dir, nodes):
        self._morph_dir = morph_dir
        self._nodes = nodes
        self._load = nm.load_neuron
        if MORPH_CACHE_SIZE is not None:
            try:
                from functools import lru_cache
            except ImportError:  # pragma: nocover
                from functools32 import lru_cache
            self._load = lru_cache(maxsize=MORPH_CACHE_SIZE)(self._load)

    def get_filepath(self, node_id):
        """ Return path to SWC morphology file corresponding to `node_id`. """
        name = self._nodes.get(node_id, 'morphology')
        return os.path.join(self._morph_dir, "%s.swc" % name)

    def get(self, node_id, transform=False):
        """
        Return NeuroM morphology object corresponding to `node_id`.

        If `transform` is True, rotate and translate morphology points
        according to `node_id` position in the circuit.
        """
        filepath = self.get_filepath(node_id)
        result = self._load(filepath)
        if transform:
            A_t = self._nodes.orientations(node_id).transpose()
            B = self._nodes.positions(node_id).values
            result = result.transform(lambda p: np.dot(p, A_t) + B)
        return result



"""
Simulation and Neural network Analysis Productivity layer.

Copyright (c) 2019, EPFL/Blue Brain Project

This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

This library is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License version 3.0 as published
by the Free Software Foundation.

This library is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
details.

You should have received a copy of the GNU Lesser General Public License
along with this library; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from bluepysnap.circuit import Circuit
from bluepysnap.exceptions import BlueSnapError



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Miscellaneous utilities.
"""

import collections
import json

import numpy as np
import six


def load_json(filepath):
    """ Load JSON from file. """
    with open(filepath) as f:
        return json.load(f)


def is_iterable(v):
    """ Check if `v` is any iterable (strings are considered scalar). """
    return isinstance(v, collections.Iterable) and not isinstance(v, six.string_types)


def ensure_list(v):
    """ Convert iterable / wrap scalar into list (strings are considered scalar). """
    if is_iterable(v):
        return list(v)
    else:
        return [v]


def euler2mat(az, ay, ax):
    """
    Build 3x3 rotation matrices from az, ay, ax rotation angles (in that order).

    Args:
        az: rotation angles around Z (Nx1 NumPy array; radians)
        ay: rotation angles around Y (Nx1 NumPy array; radians)
        ax: rotation angles around X (Nx1 NumPy array; radians)

    Returns:
        List with 3x3 rotation matrices corresponding to each of N angle triplets.

    See also:
        https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix (R = X1 * Y2 * Z3)
    """

    c1, s1 = np.cos(ax), np.sin(ax)
    c2, s2 = np.cos(ay), np.sin(ay)
    c3, s3 = np.cos(az), np.sin(az)

    mm = np.array([
        [c2 * c3, -c2 * s3, s2],
        [c1 * s3 + c3 * s1 * s2, c1 * c3 - s1 * s2 * s3, -c2 * s1],
        [s1 * s3 - c1 * c3 * s2, c3 * s1 + c1 * s2 * s3, c1 * c2],
    ])

    return [mm[..., i] for i in range(len(az))]



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Node population access.
"""

import collections

import libsonata
import numpy as np
import pandas as pd
import six

from cached_property import cached_property

from bluepysnap import utils
from bluepysnap.exceptions import BlueSnapError

DYNAMICS_PREFIX = "@dynamics:"


def _get_population_name(h5_filepath):
    populations = libsonata.NodeStorage(h5_filepath).population_names
    if len(populations) != 1:
        raise BlueSnapError(
            "Only single-population node collections are supported (found: %d)" % len(populations)
        )
    return list(populations)[0]


def _load_population(h5_filepath, csv_filepath, population):
    """
    Load node properties from SONATA Nodes.

    Returns:
        pandas.DataFrame with node properties (zero-based index).
    """
    nodes = libsonata.NodePopulation(h5_filepath, csv_filepath or '', population)

    node_count = nodes.size
    result = pd.DataFrame(index=np.arange(node_count))

    _all = libsonata.Selection([(0, node_count)])
    for attr in sorted(nodes.attribute_names):
        result[attr] = nodes.get_attribute(attr, _all)
    for attr in sorted(nodes.dynamics_attribute_names):
        result['%s%s' % (DYNAMICS_PREFIX, attr)] = nodes.get_dynamics_attribute(attr, _all)

    return result


# TODO: move to `libsonata` library
def _complex_query(prop, query):
    # pylint: disable=assignment-from-no-return
    result = np.full(len(prop), True)
    for key, value in six.iteritems(query):
        if key == '$regex':
            result = np.logical_and(result, prop.str.match(value + "\\Z"))
        else:
            raise BlueSnapError("Unknown query modifier: '%s'" % key)
    return result


def _node_ids_by_filter(nodes, props):
    """
    Return index of `nodes` rows matching `props` dict.

    `props` values could be:
        pairs (range match for floating dtype fields)
        scalar or iterables (exact or "one of" match for other fields)

    E.g.:
        >>> _node_ids_by_filter(nodes, { Node.X: (0, 1), Node.MTYPE: 'L1_SLAC' })
        >>> _node_ids_by_filter(nodes, { Node.LAYER: [2, 3] })
    """
    # pylint: disable=assignment-from-no-return
    unknown_props = set(props) - set(nodes.columns)
    if unknown_props:
        raise BlueSnapError("Unknown node properties: [{0}]".format(", ".join(unknown_props)))

    mask = np.full(len(nodes), True)
    for prop, values in six.iteritems(props):
        prop = nodes[prop]
        if issubclass(prop.dtype.type, np.floating):
            v1, v2 = values
            prop_mask = np.logical_and(prop >= v1, prop <= v2)
        elif isinstance(values, six.string_types) and values.startswith('regex:'):
            prop_mask = _complex_query(prop, {'$regex': values[6:]})
        elif isinstance(values, collections.Mapping):
            prop_mask = _complex_query(prop, values)
        else:
            prop_mask = np.in1d(prop, values)
        mask = np.logical_and(mask, prop_mask)

    return nodes.index[mask].values


class NodePopulation(object):
    """ Node population access. """

    def __init__(self, config, circuit):
        self._h5_filepath = config['nodes_file']
        self._csv_filepath = config['node_types_file']
        if 'node_sets_file' in config:
            self._node_sets = utils.load_json(config['node_sets_file'])
        else:
            self._node_sets = {}
        self._circuit = circuit

    @cached_property
    def name(self):
        """ Node population name. """
        return _get_population_name(self._h5_filepath)

    @property
    def size(self):
        """ Node population size. """
        return len(self._data)

    @property
    def property_names(self):
        """ Set of available node properties. """
        return set(self._data.columns)

    def property_values(self, prop):
        """ Set of values for a given property. """
        return set(self.get(properties=prop).unique())

    @property
    def node_sets(self):
        """ Node sets defined for this node population. """
        return self._node_sets

    @cached_property
    def _data(self):
        return _load_population(self._h5_filepath, self._csv_filepath, self.name)

    def _check_id(self, node_id):
        """ Check that single node id belongs to the circuit. """
        if node_id not in self._data.index:
            raise BlueSnapError("node id not found: %d" % node_id)

    def _check_ids(self, node_ids):
        """ Check that node ids belong to the circuit. """
        missing = pd.Index(node_ids).difference(self._data.index)
        if not missing.empty:
            raise BlueSnapError("node id not found: [%s]" % ",".join(map(str, missing)))

    def _check_property(self, prop):
        if prop not in self.property_names:
            raise BlueSnapError("No such property: '%s'" % prop)

    def ids(self, group=None, limit=None, sample=None):
        """ node ids corresponding to node `group`.

            `group` could be:
                - single node id (int)
                - list of node ids (list-like)
                - string (target name)
                - properties filter (dict-like)
                - None (i.e. 'all')

            If `sample` is specified, `sample` node ids are randomly chosen from the match result.
            If `limit` is specified, first `limit` node ids from the match result are returned.

            If `group` is a sequence (list or array), its order is preserved.
            Otherwise return result is sorted and contains no duplicates.
        """

        preserve_order = False

        if isinstance(group, six.string_types):
            if group not in self._node_sets:
                raise BlueSnapError("Undefined node set: %s" % group)
            group = self._node_sets[group]

        if group is None:
            result = self._data.index.values
        elif isinstance(group, collections.Mapping):
            result = _node_ids_by_filter(self._data, props=group)
        elif isinstance(group, np.ndarray):
            result = group
            self._check_ids(result)
            preserve_order = True
        else:
            result = utils.ensure_list(group)
            self._check_ids(result)
            preserve_order = isinstance(group, collections.Sequence)

        if sample is not None:
            result = np.random.choice(result, sample, replace=False)
            preserve_order = False
        if limit is not None:
            result = result[:limit]

        result = np.array(result, dtype=int)
        if preserve_order:
            return result
        else:
            return np.unique(result)

    def get(self, group=None, properties=None):
        """ Node properties as pandas Series / DataFrame.

            `group` could be:
                - single node id (int)
                - list of node ids (list-like)
                - string (target name)
                - properties filter (dict-like)
                - None (i.e. 'all')

            If `properties` is specified, return only selected properties (all by default).

            Returns:
                pandas Series if single node id is passed as `group`.
                pandas DataFrame indexed by node ids otherwise.
        """

        result = self._data

        if properties is not None:
            for p in utils.ensure_list(properties):
                self._check_property(p)
            result = result[properties]

        if group is not None:
            if isinstance(group, six.integer_types + (np.integer,)):
                self._check_id(group)
            else:
                group = self.ids(group)
            result = result.loc[group]

        return result

    def positions(self, group=None):
        """
        Node position(s) as pandas Series / DataFrame.

        Returns:
            pandas ('x', 'y', 'z') Series if single node id is passed as `group`.
            pandas ('x', 'y', 'z') DataFrame indexed by node ids otherwise
        """
        result = self.get(group=group, properties=['x', 'y', 'z'])
        return result.astype(float)

    def orientations(self, group=None):
        """
        Node orientation(s) as pandas Series / DataFrame.

        Returns:
            3x3 rotation matrix if single node id is passed as `group`.
            pandas Series with rotation matrices indexed by node ids otherwise
        """
        props = [
            'rotation_angle_xaxis',
            'rotation_angle_yaxis',
            'rotation_angle_zaxis',
        ]
        result = self.get(group=group, properties=props)
        if isinstance(group, six.integer_types + (np.integer,)):
            result = utils.euler2mat(
                [result['rotation_angle_zaxis']],
                [result['rotation_angle_yaxis']],
                [result['rotation_angle_xaxis']],
            )[0]
        else:
            result = pd.Series(
                utils.euler2mat(
                    result['rotation_angle_zaxis'].values,
                    result['rotation_angle_yaxis'].values,
                    result['rotation_angle_xaxis'].values,
                ),
                index=result.index,
                name='orientation'
            )
        return result

    def count(self, group=None):
        """ Total number of nodes for a given node group. """
        return len(self.ids(group))

    @cached_property
    def morph(self):
        """ Access to node morphologies. """
        from bluepysnap.morph import MorphHelper
        return MorphHelper(
            self._circuit.config['components']['morphologies_dir'],
            self
        )



# Copyright (c) 2019, EPFL/Blue Brain Project

# This file is part of BlueBrain SNAP library <https://github.com/BlueBrain/snap>

# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
BBP cell / synapse attribute namespace.
"""

from bluepysnap import edges
from bluepysnap import nodes


class Cell(object):
    """ Cell property names. """
    MORPHOLOGY = "morphology"
    ME_COMBO = "me_combo"
    MTYPE = "mtype"
    ETYPE = "etype"
    LAYER = "layer"
    REGION = "region"
    SYNAPSE_CLASS = "synapse_class"
    X = "x"
    Y = "y"
    Z = "z"
    HOLDING_CURRENT = nodes.DYNAMICS_PREFIX + 'holding_current'
    THRESHOLD_CURRENT = nodes.DYNAMICS_PREFIX + 'threshold_current'


class Synapse(object):
    """ Synapse property names. """
    PRE_GID = edges.SOURCE_NODE_ID
    POST_GID = edges.TARGET_NODE_ID

    AXONAL_DELAY = "delay"
    D_SYN = "depression_time"
    DTC = "decay_time"
    F_SYN = "facilitation_time"
    G_SYNX = "conductance"
    NRRP = "NRRP"
    TYPE = "syn_type_id"
    U_SYN = "u_syn"

    PRE_BRANCH_ORDER = "morpho_branch_order_pre"
    PRE_NEURITE_DISTANCE = "morpho_neurite_distance_pre"
    PRE_SECTION_DISTANCE = "morpho_section_distance_pre"
    PRE_SECTION_ID = "morpho_section_id_pre"

    POST_BRANCH_ORDER = "morpho_branch_order_post"
    POST_BRANCH_TYPE = "morpho_branch_type_post"
    POST_NEURITE_DISTANCE = "morpho_neurite_distance_post"
    POST_SECTION_DISTANCE = "morpho_section_distance_post"
    POST_SECTION_ID = "morpho_section_id_post"

    # presynaptic touch position (in the center of the segment)
    PRE_X_CENTER = "efferent_center_x"
    PRE_Y_CENTER = "efferent_center_y"
    PRE_Z_CENTER = "efferent_center_z"

    # presynaptic touch position (on the segment surface)
    PRE_X_SURFACE = "efferent_surface_x"
    PRE_Y_SURFACE = "efferent_surface_y"
    PRE_Z_SURFACE = "efferent_surface_z"

    # postsynaptic touch position (in the center of the segment)
    POST_X_CENTER = "afferent_center_x"
    POST_Y_CENTER = "afferent_center_y"
    POST_Z_CENTER = "afferent_center_z"

    # postsynaptic touch position (on the segment surface)
    POST_X_SURFACE = "afferent_surface_x"
    POST_Y_SURFACE = "afferent_surface_y"
    POST_Z_SURFACE = "afferent_surface_z"
